/*
 *  linux/drivers/acorn/char/i2c-s3531a.c
 *
 *  Copyright (C) 2001 Atmark Techno, Inc.
 */

#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/mc146818rtc.h>
#include <linux/init.h>

#include <linux/i2c-s3531a.h>

#define REVERSE(x) ((((x) << 7) & 0x80) | (((x) << 5) & 0x40) | \
                    (((x) << 3) & 0x20) | (((x) << 1) & 0x10) | \
                    (((x) >> 1) & 0x08) | (((x) >> 3) & 0x04) | \
                    (((x) >> 5) & 0x02) | (((x) >> 7) & 0x01))

#define CHRDEV  "i2c-s3531a"
#define VERSION CHRDEV ": i2c S-3531A driver, (C) 2001 Atmark Techno, Inc."

static struct i2c_driver s3531a_driver;

static unsigned short ignore[] = { I2C_CLIENT_END };
static unsigned short normal_addr[] = { I2C_S3531A_ADDRESS, I2C_CLIENT_END };

static struct i2c_client_address_data addr_data = {
  normal_i2c:       normal_addr,
  normal_i2c_range: ignore,
  probe:            ignore,
  probe_range:	    ignore,
  ignore:	    ignore,
  ignore_range:	    ignore,
  force:	    ignore,
};

#define DAT(x) ((unsigned int)(x->data))

static int
s3531a_attach(struct i2c_adapter *adap, int addr, unsigned short flags,
	      int kind)
{
  struct i2c_client *c;
  unsigned char buf[1];
  struct i2c_msg msgs[1] = {
    { addr | 0x1, I2C_M_RD, 1, buf }
  };
  
  c = kmalloc(sizeof(*c), GFP_KERNEL);
  if (!c)
    return -ENOMEM;
  
  strcpy(c->name, "S-3531A");
  c->id	     = s3531a_driver.id;
  c->flags   = 0;
  c->addr    = addr;
  c->adapter = adap;
  c->driver  = &s3531a_driver;
  c->data    = NULL;
  
  if (i2c_transfer(c->adapter, msgs, 1) == 1)
    DAT(c) = buf[0];
  
  return i2c_attach_client(c);
}

static int
s3531a_probe(struct i2c_adapter *adap)
{
  return i2c_probe(adap, &addr_data, s3531a_attach);
}

static int
s3531a_detach(struct i2c_client *client)
{
  i2c_detach_client(client);
  kfree(client);
  return 0;
}

static int
s3531a_get_datetime(struct i2c_client *client, struct s3531a_time *dt,
		    int datetoo)
{
  unsigned char buf[7];
  struct i2c_msg msgs[1] = {
    { client->addr | 0x2 | !datetoo, I2C_M_RD, datetoo ? 7 : 3, buf  }
  };
  int ret, i = 0;
  
  ret = i2c_transfer(client->adapter, msgs, 1);
  if (ret == 1) {
    if (datetoo) {
      dt->year = REVERSE(buf[i]);
      BCD_TO_BIN(dt->year);
      i++;
      dt->mon  = REVERSE(buf[i]);
      BCD_TO_BIN(dt->mon);
      i++;
      dt->mday = REVERSE(buf[i]);
      BCD_TO_BIN(dt->mday);
      i++;
      dt->wday = REVERSE(buf[i]);
      BCD_TO_BIN(dt->wday);
      i++;
    }
    dt->hour = REVERSE(buf[i]) & 0x3f;
    BCD_TO_BIN(dt->hour);
    i++;
    dt->min  = REVERSE(buf[i]);
    BCD_TO_BIN(dt->min);
    i++;
    dt->sec  = REVERSE(buf[i]);
    BCD_TO_BIN(dt->sec);
    
    ret = 0;
  }
  
  return ret;
}

static int
s3531a_set_datetime(struct i2c_client *client, struct s3531a_time *dt,
		    int datetoo)
{
  unsigned char buf[7];
  struct i2c_msg msgs[1] = {
    { client->addr | 0x2 | !datetoo, 0, datetoo ? 7 : 3, buf  }
  };
  int ret, i = 0;
  
  if (datetoo) {
    buf[i] = dt->year;
    BIN_TO_BCD (buf[i]);
    buf[i] = REVERSE(buf[i]);
    i++;
    buf[i] = dt->mon;
    BIN_TO_BCD (buf[i]);
    buf[i] = REVERSE(buf[i]);
    i++;
    buf[i] = dt->mday;
    BIN_TO_BCD (buf[i]);
    buf[i] = REVERSE(buf[i]);
    i++;
    buf[i] = dt->wday;
    BIN_TO_BCD (buf[i]);
    buf[i] = REVERSE(buf[i]);
    i++;
  }
  buf[i] = dt->hour;
  BIN_TO_BCD (buf[i]);
  buf[i] = REVERSE(buf[i]);
  i++;
  buf[i] = dt->min;
  BIN_TO_BCD (buf[i]);
  buf[i]   = REVERSE(buf[i]);
  i++;
  buf[i] = dt->sec;
  BIN_TO_BCD (buf[i]);
  buf[i] = REVERSE(buf[i]);
  
  ret = i2c_transfer(client->adapter, msgs, 1);
  if (ret == 1) {
    ret = 0;
  }
  
  return ret;
}

static int
s3531a_get_ctrl(struct i2c_client *client, unsigned char *ctrl)
{
  *ctrl = DAT(client);
  return 0;
}

static int
s3531a_set_ctrl(struct i2c_client *client, unsigned char *ctrl)
{
  unsigned char buf[1];
  struct i2c_msg msgs[1] = {
    { client->addr | 0x1, 0, 1, buf }
  };
  int ret;
  
  buf[0] = *ctrl;
  ret = i2c_transfer(client->adapter, msgs, 1);
  if (ret == 1) {
    msgs[0].flags = I2C_M_RD;
    ret = i2c_transfer(client->adapter, msgs, 1);
    if (ret == 1) {
      DAT(client) = buf[0];
      ret = 0;
    }
  }
  
  return ret;
}

static int
s3531a_command(struct i2c_client *client, unsigned int cmd, void *arg)
{
  switch (cmd) {
  case RTC_GETTIME:
    return s3531a_get_datetime(client, arg, 0);
    
  case RTC_GETDATETIME:
    return s3531a_get_datetime(client, arg, 1);
    
  case RTC_SETTIME:
    return s3531a_set_datetime(client, arg, 0);
    
  case RTC_SETDATETIME:
    return s3531a_set_datetime(client, arg, 1);
    
  case RTC_GETCTRL:
    return s3531a_get_ctrl(client, arg);
    
  case RTC_SETCTRL:
    return s3531a_set_ctrl(client, arg);
  }

  return -EINVAL;
}

static struct i2c_driver s3531a_driver = {
  name:		  "S-3531A",
  id:		  I2C_DRIVERID_S3531A,
  flags:	  I2C_DF_NOTIFY,
  attach_adapter: s3531a_probe,
  detach_client:  s3531a_detach,
  command:	  s3531a_command
};

static __init int s3531a_init(void)
{
  int ret;
  
  ret = i2c_add_driver(&s3531a_driver);
  if (ret) {
    printk (KERN_WARNING CHRDEV ": Unable to register %s driver.\n",
	    s3531a_driver.name);
    return ret;
  }
  
  printk (KERN_INFO VERSION "\n");
  return 0;
}

static __exit void s3531a_cleanup(void)
{
  i2c_del_driver(&s3531a_driver);
}

module_init(s3531a_init);
module_exit(s3531a_cleanup);
